/*<vision.cpp: Functions for object parts' feature extraction>
    Copyright (C) <2017>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "vision.h"

vision::vision()
{
}


vision::~vision()
{
}

CloudPtr vision::segment_pcd_euc(CloudPtr cloud)
{
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
  tree->setInputCloud (cloud);

  std::vector<pcl::PointIndices> cluster_indices;
  pcl::EuclideanClusterExtraction<pcl::PointXYZ> ec;
  ec.setClusterTolerance (0.02); // 2cm
  //ec.setMinClusterSize (1000);
	ec.setMinClusterSize (500);
  //ec.setMaxClusterSize (25000);
  ec.setSearchMethod (tree);
  ec.setInputCloud (cloud);
  ec.extract (cluster_indices);  

	CloudPtr cloud_c(new Cloud);
  int j = 0;
  for(std::vector<pcl::PointIndices>::const_iterator it = cluster_indices.begin (); it != cluster_indices.end (); ++it)
  {
    for (std::vector<int>::const_iterator pit = it->indices.begin (); pit != it->indices.end (); ++pit)
      cloud_c->points.push_back (cloud->points[*pit]); //*
		j++;
  }

  if(debug_)
  {
    pcl::visualization::PCLVisualizer viewer("segmented cloud from background");
    viewer.addPointCloud(cloud_c);
    viewer.spin();
  }

	return cloud_c;
}

CloudPtr vision::segment_pcd_euc_iter(CloudPtr cloud)
{
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
  tree->setInputCloud (cloud);

  pcl::SACSegmentation<pcl::PointXYZ> seg;
  pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
  pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);

  seg.setOptimizeCoefficients (true);
  seg.setModelType (pcl::SACMODEL_PLANE);
  seg.setMethodType (pcl::SAC_RANSAC);
  seg.setMaxIterations (100);
  seg.setDistanceThreshold (0.02);

	int i=0, nr_points = cloud->points.size ();
	CloudPtr cloud_filtered(new Cloud);
	*cloud_filtered = *cloud;
	CloudPtr cloud_f(new Cloud);
  while (cloud_filtered->points.size () > 0.4 * nr_points)
  {
    // Segment the largest planar component from the remaining cloud
    seg.setInputCloud (cloud_filtered);
    seg.segment (*inliers, *coefficients);
    if (inliers->indices.size () == 0)
    {
      std::cout << "Could not estimate a planar model for the given dataset." << std::endl;
      break;
    }

    // Extract the planar inliers from the input cloud
    pcl::ExtractIndices<pcl::PointXYZ> extract;
    extract.setInputCloud (cloud_filtered);
    extract.setIndices (inliers);
   // extract.setNegative (false);

    // Get the points associated with the planar surface
   // extract.filter (*cloud_plane);
    //std::cout << "PointCloud representing the planar component: " << cloud_plane->points.size () << " data points." << std::endl;

    // Remove the planar inliers, extract the rest
    extract.setNegative (true);
    extract.filter (*cloud_f);
    *cloud_filtered = *cloud_f;
  }

  std::vector<pcl::PointIndices> cluster_indices;
  pcl::EuclideanClusterExtraction<pcl::PointXYZ> ec;
  ec.setClusterTolerance (0.02); // 2cm
  //ec.setMinClusterSize (1000);
	ec.setMinClusterSize (500);
  //ec.setMaxClusterSize (25000);
  ec.setSearchMethod (tree);
  ec.setInputCloud (cloud_filtered);
  ec.extract (cluster_indices);  


	CloudPtr cloud_c(new Cloud);
  int j = 0;
  for(std::vector<pcl::PointIndices>::const_iterator it = cluster_indices.begin (); it != cluster_indices.end (); ++it)
  {
    for (std::vector<int>::const_iterator pit = it->indices.begin (); pit != it->indices.end (); ++pit)
      cloud_c->points.push_back (cloud_filtered->points[*pit]); //*
		j++;
  }

  if(debug_)
  {
    pcl::visualization::PCLVisualizer viewer("segmented cloud from background");
    viewer.addPointCloud(cloud_c);
    viewer.spin();
  }

	return cloud_c;
}

CloudPtr vision::load_cloud(string file_name)
{
  pcl::PCDReader reader;
  CloudPtr cloud (new Cloud);
  reader.read (file_name, *cloud);

	file_name_ = file_name.substr(file_name.find_last_of("/")+1);
	file_name_ = file_name_.substr(0,file_name_.find(".pcd"));
/*
	if(debug_)
	{
		pcl::visualization::PCLVisualizer viewer("Input pcd");
		viewer.addPointCloud(cloud);
		viewer.spin();  
	}
*/
  return cloud;
}

vector<CloudPtr> vision::read_pcd_parts(string dir_name,CloudPtr &cloud_orig,vector<int> &aff_ids)
{
	dir_name_ = dir_name;
	vector<string> files = get_all_files(dir_name,"_part.pcd");
	vector<string> files_labels = get_all_files(dir_name,"_l.txt");
	vector<CloudPtr> parts(files.size());
	aff_ids = vector<int>(files.size());
	for(size_t i = 0; i < files.size(); i++)
	{
		CloudPtr cloud(new Cloud);
		cloud = load_cloud(files[i]);
		int id;

		string part_name = files[i].substr(files[i].find_last_of("/")+1);
		part_name = part_name.substr(0,part_name.find("_part"));
		part_name = part_name.substr(part_name.find_last_of("_")+1);

		stringstream ss(part_name);
		ss >> id;	
		cout << "part " << id << endl;
		id --;
		parts[id] = cloud;

		string l_name = files[i];
		l_name = l_name.substr(0,l_name.find("_part"));	
		l_name = l_name + "_l.txt";
		ifstream ifs(l_name.c_str());
		string line;
		if(ifs.good())
			getline(ifs,line);
		stringstream ss_l(line);
		int id_l;
		ss_l >> id_l;
		cout << "part affordance label: " << id_l << endl;
		aff_ids[id] = id_l;
	}
	cout << "#of parts: " << parts.size() << endl;
	files = get_all_files(dir_name,"_pcd_all.pcd");
	cloud_orig = load_cloud(files[0]);
	cloud_orig = segment_pcd_euc(cloud_orig);

	vector<CloudPtr> cloud_parts_new;
	vector<int> aff_ids_new;

	if(cloud_orig->points.size() > 0)
	{
		search::KdTree<PointType>::Ptr tree (new search::KdTree<PointType>);
		tree->setInputCloud(cloud_orig);
		vector<int> k_indices;
		vector<float> k_sqr_distances;
		int k = 2;

		for(size_t i = 0; i < parts.size(); i++)
		{
			CloudPtr cloud_part = parts[i];
			if(cloud_part->points.size() < 10)
				continue;
			CloudPtr cloud_n(new Cloud);
			for(size_t ip = 0; ip < cloud_part->points.size(); ip++)
			{
				tree->nearestKSearch(*cloud_part,ip,k,k_indices,k_sqr_distances);
				if(k_sqr_distances[0] <= 1e-5)
					cloud_n->points.push_back(cloud_part->points[ip]);
			}
			*parts[i] = *cloud_n;	
			if(cloud_n->points.size() > 10)
			{
				cloud_parts_new.push_back(cloud_n);
				aff_ids_new.push_back(aff_ids[i]);
			}

		}
	}
	parts = cloud_parts_new;
	aff_ids = aff_ids_new;

	file_name_ = files[0].substr(files[0].find_last_of("/")+1);
	file_name_ = file_name_.substr(0,file_name_.find("_pcd_all"));
	
	return parts;
}

vector<string> vision::get_all_files(string dir_name,string pattern)
{
	int min_length = 1;
  vector<string> files;
  DIR *dir;
  struct dirent *ent;
  if ((dir = opendir (dir_name.c_str())) != NULL) 
  {
    while ((ent = readdir (dir)) != NULL) {
      string name = ent->d_name;
			if( name.length() < min_length )
				continue;
      if( name.find(pattern) == string::npos )
        continue;
      string file_name = dir_name + name;
      files.push_back(file_name);
    }
    closedir (dir);
  } 
  else
     cout << "can't open dir: " << dir_name <<endl;

  return files;
}

void vision::extract_between_patch_features()
{
	string dir_path = parts_path;
	//dir_path = "/media/Seagate\ Expansion\ Drive/3dv-files/parts-pcd-fg/";
  vector<string> dir_all = get_all_files(dir_path,"");  
  for(size_t i = 0; i < dir_all.size(); i++)
  {
    if(dir_all[i].find(".") != string::npos)
      continue;
    string cur_dir = dir_all[i] + "/";
    cout << cur_dir << endl;
    string file_name = cur_dir + "segs.txt";
    map<int,set<int> > segs_nns;
    ifstream ifs(file_name.c_str());
    while(ifs.good())
    {
      string line;
      getline(ifs,line);
      if(line.length() < 2)
        continue;
      vector<double> vals;
      while(line.find(" ") != string::npos)
      { 
        double val;
        string tmp = line.substr(0,line.find_first_of(" "));
        line = line.substr(line.find_first_of(" ")+1);
        stringstream ss(tmp);
        ss >> val;
        vals.push_back(val);
      }
      stringstream ss(line);
      double val;
      ss >> val;
      vals.push_back(val);  
      if(vals.size() <2)
        continue;
      //cout << "#of nns for " << vals[0] << " : " << vals.size() -1 << endl;
      for(size_t j = 1; j < vals.size(); j++)
        segs_nns[vals[0]].insert(vals[j]);
    }
    map<int,CloudPtr> seg_map;
    vector<string> pcd_files = get_all_files(cur_dir,".pcd"); 
    string file_part_name = cur_dir + "part.txt";
    ofstream ofp(file_part_name.c_str());
    for(size_t ip = 0; ip < pcd_files.size(); ip++)
    {
      string cur_pcd_name = pcd_files[ip];
      cur_pcd_name = cur_pcd_name.substr(cur_pcd_name.find_last_of("_")+1);
      cur_pcd_name = cur_pcd_name.substr(0,cur_pcd_name.find_last_of("."));
      stringstream ssp(cur_pcd_name);
      int id_p;
      ssp >> id_p;

      string part_id_str = pcd_files[ip].substr(pcd_files[ip].find_last_of("/")+1);
      part_id_str = part_id_str.substr(part_id_str.find_first_of("_")+1);
      part_id_str = part_id_str.substr(0,part_id_str.find_first_of("_"));
      stringstream ss_p(part_id_str);
      int part_id;
      ss_p >> part_id;
      
      ofp << part_id << " " << id_p << endl;
      CloudPtr cur_cloud(new Cloud);
      cur_cloud = load_cloud(pcd_files[ip]); 
      seg_map[id_p] = cur_cloud;  
    //  cout << "id for " << pcd_files[ip] << " : " << id_p << endl;
    }
    ofp.close();

    string file_cvt_name = cur_dir + "cvt.txt";
    ofstream ofc(file_cvt_name.c_str());
    map<int,set<int> >::iterator it;
    for(it = segs_nns.begin(); it != segs_nns.end(); it++)
    { 
      CloudPtr cloud1(new Cloud);
      cloud1 = seg_map[(*it).first];      
      set<int>::iterator its;
      set<int> nns = (*it).second;
      for(its = nns.begin(); its != nns.end(); its++)
      {
        CloudPtr cloud2(new Cloud);
        cloud2 = seg_map[*its];
        double cvt = compute_curvature_between_patches(cloud1,cloud2);
        if( isnan(fabs(cvt)) || isinf(fabs(cvt)) )
          cvt = 0;
   //     cout << "curvature: " << cvt << endl;
        ofc << (*it).first << " " << (*its) << " " << cvt << endl;
      }
    }
    ofc.close();
  }

}

double vision::compute_curvature_between_patches(CloudPtr cloud1,CloudPtr cloud2)
{
  CloudPtr cloud(new Cloud);
  for(size_t ip = 0; ip < cloud1->points.size(); ip++)
    cloud->points.push_back(cloud1->points[ip]);
  for(size_t ip = 0; ip < cloud2->points.size(); ip++)
    cloud->points.push_back(cloud2->points[ip]);

//  cout << "#of points: " << cloud->points.size() << endl;
  CloudPtr cloud_cur(new Cloud);
  vector<int> idx;
  set<int> idx_set;

  search::KdTree<PointType>::Ptr tree (new search::KdTree<PointType>);
  tree->setInputCloud(cloud);
  vector<int> k_indices;
  vector<float> k_sqr_distances;
  double radius = 0.01;

  for(size_t ip = 0; ip < cloud1->points.size(); ip++)
  {
    tree->radiusSearch(*cloud,ip,radius,k_indices,k_sqr_distances);
//    tree->nearestKSearch(*cloud,ip,30,k_indices,k_sqr_distances);
    bool found = false;
    for(size_t ik = 0;( (!found) && (ik < k_indices.size()) ); ik++)
    {
      if( k_indices[ik] >= cloud1->points.size() )
			{
        found = true;
			}
    } 
    if(!found)
      continue;
    idx_set.insert(ip);
    for(size_t ik = 0; ik < k_indices.size(); ik++)
      idx_set.insert(k_indices[ik]); 
  }

  set<int>::iterator its;
  for(its = idx_set.begin(); its != idx_set.end(); its++)
    idx.push_back(*its);
  NormalEstimation<PointType, pcl::Normal> ne;
  ne.setInputCloud(cloud);
  ne.setSearchMethod(tree);
  ne.setRadiusSearch (0.01);
  float nx,ny,nz,cvt;
  ne.computePointNormal(*cloud,idx,nx,ny,nz,cvt);  
  return cvt;
}

void vision::learn_parts_features_pcl_test()
{
  vector<string> dir_all = get_all_files(files_path_,".pcd");

  for(size_t i = 0; i < dir_all.size(); i++)
  {
    cout << "to read data for: " << dir_all[i] << endl;
    CloudPtr cloud(new Cloud);
		cloud = load_cloud(dir_all[i]);
		vector<int> indices;
		pcl::removeNaNFromPointCloud(*cloud,*cloud,indices);
		cloud = segment_pcd_euc(cloud);

		PointCloud<pcl::Normal>::Ptr normals = estimate_normal(cloud);
    map<int,set<int> > segs_nns;
    map<int,CloudPtr> segment_voxels;
    map<int,PointCloud<pcl::Normal>::Ptr> segment_normals;      
    vector<pcl::PointIndices> seg_ids = get_regions_from_pcd(cloud,normals,segment_voxels,segment_normals,segs_nns);
    map<int,vec> hist_sis;
    map<int,vec> hist_cvs;
    map<int,vec> fet_si,fet_cv,fet_si_max,fet_cv_max;
    extract_features_patches(segment_voxels,segment_normals,segs_nns,hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max);
    save_patch_test_features(hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max,segment_voxels,segs_nns);
  }

}

void vision::learn_parts_features_pcl_test_planar_seg()
{
  vector<string> dir_all = get_all_files(files_path_,".pcd");

  for(size_t i = 0; i < dir_all.size(); i++)
  {
    cout << "to read data for: " << dir_all[i] << endl;
    CloudPtr cloud(new Cloud);
		cloud = load_cloud(dir_all[i]);
		vector<int> indices;
		pcl::removeNaNFromPointCloud(*cloud,*cloud,indices);
		cloud = segment_pcd_euc_iter(cloud);

		PointCloud<pcl::Normal>::Ptr normals = estimate_normal(cloud);
    map<int,set<int> > segs_nns;
    map<int,CloudPtr> segment_voxels;
    map<int,PointCloud<pcl::Normal>::Ptr> segment_normals;      
    vector<pcl::PointIndices> seg_ids = get_regions_from_pcd(cloud,normals,segment_voxels,segment_normals,segs_nns);
    map<int,vec> hist_sis;
    map<int,vec> hist_cvs;
    map<int,vec> fet_si,fet_cv,fet_si_max,fet_cv_max;
    extract_features_patches(segment_voxels,segment_normals,segs_nns,hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max);
    save_patch_test_features(hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max,segment_voxels,segs_nns);
  }

}

void vision::save_patch_test_features(map<int,vec> hist_sis,map<int,vec> hist_cvs,map<int,vec> fet_si,map<int,vec> fet_cv,map<int,vec> fet_si_max,map<int,vec> fet_cv_max,map<int,CloudPtr> segment_voxels,map<int,set<int> > segs_nns)
{
 // string path_fets_txt;
	//path_fets_txt = default_path_ + "patches-pcl-scenes-txt/";
  boost::filesystem::path fet_path_txt(path_fets_txt);
  boost::filesystem::create_directories(fet_path_txt);

  //string parts_path;
	//parts_path = default_path_ + "parts-scenes-pcd/";
  boost::filesystem::path parts_path_dir(parts_path);
  boost::filesystem::create_directories(parts_path_dir);

  string obj_path = parts_path + file_name_ + "/";
  boost::filesystem::path obj_path_dir(obj_path);
  boost::filesystem::create_directories(obj_path_dir);
 
  string nn_file_name = obj_path + "segs.txt";
  ofstream ofs(nn_file_name.c_str());
	map<int,CloudPtr>::iterator itc;
	for(itc = segment_voxels.begin(); itc != segment_voxels.end(); itc++)
	{
		ofs << (*itc).first;
		if(segs_nns.find((*itc).first) == segs_nns.end()) 
    	ofs << endl;		
		else
		{
    	set<int>::iterator its;
			set<int> nns = segs_nns[(*itc).first];
    	for(its = nns.begin(); its !=nns.end(); its++)
      	ofs << " " << (*its);
    	ofs << endl;			
		}
	}


  map<int,set<int> >::iterator itn;
/*  for(itn = segs_nns.begin(); itn != segs_nns.end(); itn++)
  {
    ofs << (*itn).first;
    set<int> nns = (*itn).second;
    set<int>::iterator its;
    for(its = nns.begin(); its !=nns.end(); its++)
      ofs << " " << (*its);
    ofs << endl;
  }
*/
  ofs.close();
	
  string file_cvt_name = obj_path + "cvt.txt";
  ofstream ofc(file_cvt_name.c_str());
  for(itn = segs_nns.begin(); itn != segs_nns.end(); itn++)
  { 
  	CloudPtr cloud1(new Cloud);
    cloud1 = segment_voxels[(*itn).first];      
		set<int>::iterator its;
    set<int> nns = (*itn).second;
    for(its = nns.begin(); its != nns.end(); its++)
    {
    	CloudPtr cloud2(new Cloud);
      cloud2 = segment_voxels[*its];
      double cvt = compute_curvature_between_patches(cloud1,cloud2);
      if( isnan(fabs(cvt)) || isinf(fabs(cvt)) )
      	cvt = 0;
      ofc << (*itn).first << " " << (*its) << " " << cvt << endl;
    }
  }
  ofc.close();

	map<int,vec>::iterator it_f;
	for(it_f = hist_sis.begin(); it_f != hist_sis.end(); it_f++)
	{
		int s_id = (*it_f).first;
//		cout << "to save: "<< s_id << " : " << hist_sis.size() << endl;
		stringstream ss_t;
		ss_t << s_id;
		string st = ss_t.str();
    string si_name,cv_name,sim_name,cvm_name,root_name;		

    root_name = file_name_ + "_" + st + "_sim";
    si_name = path_fets_txt + root_name + ".txt";
    fet_si[s_id].save(si_name,arma_ascii);

    root_name = file_name_ + "_" + st + "_cvm";
    cv_name = path_fets_txt + root_name + ".txt"; 
  	fet_cv[s_id].save(cv_name,arma_ascii);

    root_name = file_name_ + "_" + st + "_simax"; 
    sim_name = path_fets_txt + root_name + ".txt";
    fet_si_max[s_id].save(sim_name,arma_ascii);
	
    root_name = file_name_ + "_" + st + "_cvmax";
    cvm_name = path_fets_txt + root_name + ".txt";
    fet_cv_max[s_id].save(cvm_name,arma_ascii);

    root_name = file_name_ + "_" + st + "_histn";
    si_name = path_fets_txt + root_name + ".txt";
		hist_normals_[s_id].save(si_name,arma_ascii);

    root_name = file_name_ + "_" + st + "_histc";
    si_name = path_fets_txt + root_name + ".txt";
    hist_cvts_[s_id].save(si_name,arma_ascii);

		vec cvt_all(cvts_all_[s_id].size());	
		cvt_all.fill(0);
		for(size_t ie = 0; ie < cvts_all_[s_id].size(); ie++)
			cvt_all(ie) = cvts_all_[s_id][ie];
    root_name = file_name_ + "_" + st + "_curv";
    si_name = path_fets_txt + root_name + ".txt";
    cvt_all.save(si_name,arma_ascii);

		Mat<double> normal_all(normals_all_[s_id].size(),3);
		normal_all.fill(0);
		for(size_t ie = 0; ie < normals_all_[s_id].size(); ie++)
		{	
			normal_all(ie,0) = normals_all_[s_id][ie](0);
			normal_all(ie,1) = normals_all_[s_id][ie](1);
			normal_all(ie,2) = normals_all_[s_id][ie](2);
		}						
    root_name = file_name_ + "_" + st + "_normal";
    si_name = path_fets_txt + root_name + ".txt";
    normal_all.save(si_name,arma_ascii);

    string cloud_name = obj_path + "cls" + "_" + st + ".pcd";
		CloudPtr cur_cloud = segment_voxels[s_id];
		cur_cloud->width = 1;
		cur_cloud->height = cur_cloud->points.size();
		pcl::io::savePCDFileASCII(cloud_name,*cur_cloud);
	}

}

void vision::learn_parts_features_pcl(string d_path)
{
	vector<string> dir_all = get_all_files(files_path_,"");

  for(size_t i = 0; i < dir_all.size(); i++)
  {
    if(dir_all[i].find(".") != string::npos)
      continue;

    string cur_dir = dir_all[i] + "/";

    cout << "to read data for: " << cur_dir << endl;
    CloudPtr cloud(new Cloud);

    vector<int> aff_ids;
    vector<CloudPtr> parts = read_pcd_parts(cur_dir,cloud,aff_ids);
		if(cloud->points.size() < 10)
			continue;

		cout << "to estimate normals" << endl;
    PointCloud<pcl::Normal>::Ptr normals = estimate_normal(cloud);
    map<int,set<int> > segs_nns;
    map<int,CloudPtr> segment_voxels;
		cout << "to get regions" << endl;
    map<int,PointCloud<pcl::Normal>::Ptr> segment_normals;      
    vector<pcl::PointIndices> seg_ids = get_regions_from_pcd(cloud,normals,segment_voxels,segment_normals,segs_nns);
		cout << "after get regions" << endl;
    map<int,vector<int> > p_ids = get_segments_parts_ids(segment_voxels,parts,cloud);
    map<int,vec> hist_sis;
    map<int,vec> hist_cvs;
    map<int,vec> fet_si,fet_cv,fet_si_max,fet_cv_max;
		cout << "to get features" << endl;
    extract_features_patches(segment_voxels,segment_normals,segs_nns,hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max);
    aggregate_part_features(p_ids,aff_ids,hist_sis,hist_cvs,fet_si,fet_cv,fet_si_max,fet_cv_max,segment_voxels,segs_nns);
  }

}

map<int,vector<int> > vision::get_segments_parts_ids(map<int,CloudPtr> segment_voxels,vector<CloudPtr> parts,CloudPtr cloud)
{
	map<int,vector<int> > p_ids;

	map<int,CloudPtr>::iterator it;
  search::KdTree<PointType>::Ptr tree (new search::KdTree<PointType>);

	for(it = segment_voxels.begin(); it != segment_voxels.end(); it++)
	{
		double min_dist = FLT_MAX;
		int min_id = -1;
		vector< int > indices;
		for(size_t ip = 0; ip < (*it).second->points.size();ip++)
			indices.push_back(ip);
		for(size_t i = 0; i < parts.size(); i++)
		{
			tree->setInputCloud(parts[i]);
			vector<vector<int> > k_indices;
			vector<vector<float> > k_sqr_distances;
			tree->nearestKSearch(*(*it).second,indices,2,k_indices,k_sqr_distances);
			double cur_dist = 0;
			for(size_t i1 = 0; i1 < indices.size(); i1++)
				cur_dist += k_sqr_distances[i1][0];
			if(cur_dist < min_dist)
			{
				min_dist = cur_dist;
				min_id = i;
			}
		}
		if(min_id < 0)
			continue;
		p_ids[min_id].push_back((*it).first);	
	}
/*
	map<int,vector<int> >::iterator itp;
	for(itp = p_ids.begin(); itp != p_ids.end(); itp++)
	{
		int id = (*itp).first;
		vector<int> segs_ids = (*itp).second;
		cout << "part: " << id << endl;
		for(size_t i = 0; i < segs_ids.size(); i++)
			cout << segs_ids[i] << endl;
		cout << endl;
	}
*/
	return p_ids;
}

PointCloud<pcl::Normal>::Ptr vision::estimate_normal(CloudPtr cloud)
{
  NormalEstimation<PointType, pcl::Normal> ne;
  ne.setInputCloud(cloud);
  pcl::search::KdTree<PointType>::Ptr tree (new pcl::search::KdTree<PointType> ());
  ne.setSearchMethod(tree);
  PointCloud<pcl::Normal>::Ptr cloud_normals (new PointCloud<pcl::Normal>);
//  ne.setKSearch(50);
  ne.setRadiusSearch(0.01);
  ne.setSearchSurface(cloud);
  ne.compute (*cloud_normals);
  return cloud_normals;
}

vector<pcl::PointIndices> vision::get_regions_from_pcd(CloudPtr cloud,PointCloud<pcl::Normal>::Ptr normals,map<int,CloudPtr> &segment_voxels,map<int,PointCloud<pcl::Normal>::Ptr> &segment_normals,map<int,set<int> > &segs_nns,double seg_thresh)
{
  pcl::RegionGrowing<PointType,pcl::Normal> reg;
  pcl::search::Search<PointType>::Ptr tree = boost::shared_ptr<pcl::search::Search<PointType> > (new pcl::search::KdTree<PointType>);
  reg.setMinClusterSize(100);
  reg.setMaxClusterSize(1000000);
  reg.setSearchMethod(tree);
  reg.setNumberOfNeighbours(30);
  reg.setInputCloud (cloud);
  reg.setInputNormals (normals);
	if(seg_thresh < 1e-5)
		seg_thresh = smooth_thresh_;
  reg.setSmoothnessThreshold (seg_thresh / 180.0 * M_PI);
  reg.setCurvatureThreshold (1.0);
  vector<pcl::PointIndices> clusters;
  reg.extract (clusters);


  //cout << "#of clusters: " << clusters.size() << endl;

	set<int> all_cls_ids;
	map<int,int> cls_pts_map;
	CloudPtr cloud_ref(new Cloud);
	map<int,int> pts_ref_map;

	if(clusters.size() == 0)
	{
		pcl::PointIndices p_idx;
		for(size_t ip = 0; ip < cloud->points.size(); ip++)
			p_idx.indices.push_back(ip);
		clusters.push_back(p_idx);
	}

	for(size_t i = 0; i < clusters.size(); i++)
	{
		vector<int> ids_cur = clusters[i].indices;
		for(size_t ip = 0; ip < ids_cur.size(); ip++)
		{
			all_cls_ids.insert(ids_cur[ip]);
			cls_pts_map[ids_cur[ip]] = i;
			cloud_ref->points.push_back(cloud->points[ids_cur[ip]]);
			pts_ref_map[cloud_ref->points.size()-1] = ids_cur[ip];
		}
	}
//	cout << "#pts: " << cloud->points.size() << " : " << cloud_ref->points.size() << " : " << clusters.size() << endl;
	search::KdTree<PointType>::Ptr tree_pts (new search::KdTree<PointType>);
	tree_pts->setInputCloud(cloud_ref);
	//tree_pts->setInputCloud(cloud);

	map<int,int> cls_npts_map;	 
	double min_dist = 1e-4;
	bool is_changed = true;
//	while(all_cls_ids.size() < cloud->points.size())
	while(is_changed)
	{
		vector<int> changed;
		for(size_t i = 0; i < cloud->points.size(); i++)
		{
			if(all_cls_ids.find(i) != all_cls_ids.end())
				continue;
			
			vector<int> k_indices;
			int k = 2;
  		vector<float> k_sqr_distances;	
			tree_pts->nearestKSearch(*cloud,i,k,k_indices,k_sqr_distances);
	/*		if(all_cls_ids.find(k_indices[0]) == all_cls_ids.end())
				continue;
*/		
			if( (k_sqr_distances[0] > min_dist) && (is_changed) )
				continue;

			int min_id = pts_ref_map[k_indices[0]];
			cls_npts_map[i] = cls_pts_map[min_id];
			changed.push_back(i);
		}		
		
		for(size_t i = 0; i < changed.size(); i++)
		{
			all_cls_ids.insert(changed[i]);	
			cloud_ref->points.push_back(cloud->points[changed[i]]);
			pts_ref_map[cloud_ref->points.size()-1] = changed[i];
			cls_pts_map[changed[i]] = cls_npts_map[changed[i]];
		}
		
		tree_pts->setInputCloud(cloud_ref);
/*
		if(changed.size() == 0)
			min_dist = 5e-3;
*/
		

		if(changed.size() == 0)
			is_changed = false;
		else
			is_changed = true;


		//cout << "changed: " << is_changed << endl;
	}

/*
	for(size_t i = 0; i < cloud->points.size(); i++)
	{
		if(all_cls_ids.find(i) != all_cls_ids.end())
			continue;
			
		vector<int> k_indices;
		int k = 2;
  	vector<float> k_sqr_distances;	
		tree_pts->nearestKSearch(*cloud,i,k,k_indices,k_sqr_distances);
		//cout << "after search" << endl;
		int min_id = pts_ref_map[k_indices[0]];

		cls_npts_map[i] = cls_pts_map[min_id];
	}
*/
	//cout << "to assign clusters" << endl;
	map<int,int>::iterator itm;
	for(itm = cls_npts_map.begin(); itm != cls_npts_map.end(); itm++)
	{
		int cls_id = (*itm).second;
		int p_id = (*itm).first;
		clusters[cls_id].indices.push_back(p_id);
	}
	//cout << "after point assign" << endl;
	ColorCloudPtr cloud_deb(new ColorCloud);
	vector<Eigen::Vector3f> colors;
	for(size_t i = 0; i < clusters.size(); i++)
	{
		Eigen::Vector3f c_vec;
		c_vec(0) = rand() % 255; c_vec(1) = rand() % 255; c_vec(2) = rand() % 255; 
		colors.push_back(c_vec);
	}

  for(size_t i = 0; i < clusters.size(); i++)
  {
    CloudPtr cloud_cur(new Cloud);
    PointCloud<pcl::Normal>::Ptr normal_cur(new PointCloud<pcl::Normal>);
    vector<int> ids_cur = clusters[i].indices;
    for(size_t ip = 0; ip < ids_cur.size(); ip++)
    {
      cloud_cur->points.push_back(cloud->points[ids_cur[ip]]); 
      normal_cur->points.push_back(normals->points[ids_cur[ip]]);

			ColorPointType p;
			p.x = cloud->points[ids_cur[ip]].x; p.y = cloud->points[ids_cur[ip]].y; p.z = cloud->points[ids_cur[ip]].z;
			p.r = colors[i](0); p.g = colors[i](1); p.b = colors[i](2);
			cloud_deb->points.push_back(p);
    }
    segment_voxels[i] = cloud_cur;
    segment_normals[i] = normal_cur;
  }

  map<int,CloudPtr>::iterator it;
	double dist_thresh = 1e-3;
  for(it = segment_voxels.begin(); it != segment_voxels.end(); it++)
  {
    CloudPtr cloud_orig = (*it).second;
    int id1 = (*it).first;
    map<int,CloudPtr>::iterator it2;
    for(it2 = segment_voxels.begin(); it2 != segment_voxels.end(); it2++)   
    {
      int id2 = (*it2).first;
      if(id1 == id2) 
        continue;
      if(segs_nns[id1].find(id2) != segs_nns[id1].end()) 
        continue;
      CloudPtr cloud2 = (*it2).second;
      CloudPtr cloud_cur(new Cloud);
      *cloud_cur = *cloud_orig;
      for(size_t ip = 0; ip < cloud2->points.size(); ip++)  
        cloud_cur->points.push_back(cloud2->points[ip]);
      search::KdTree<PointType>::Ptr tree (new search::KdTree<PointType>);
      tree->setInputCloud(cloud_cur); 
      vector<int> k_indices;
      vector<float> k_sqr_distances;
			int k = 10;
      double radius = 0.005;
			dist_thresh = radius;
			int num_vote = 0;
      bool found = false;
			
			for(size_t ip = 0;ip < cloud_orig->points.size(); ip++)
      {
				bool cur_found = false;
				tree->nearestKSearch(*cloud_orig,ip,k,k_indices,k_sqr_distances);
        for(int ik = 0; ik < k_indices.size(); ik++)
        {
         	if(k_indices[ik] >= cloud_orig->points.size()) 
					{
            found = true;
						cur_found = true;
					}
        }        
    		if(cur_found)
					num_vote++;	
      }
			
      if(found)
      {
        segs_nns[id1].insert(id2);
        segs_nns[id2].insert(id1);
      }
    }
  }

  ColorCloudPtr colored_cloud = reg.getColoredCloud ();
	if(debug_)
	{
		pcl::visualization::PCLVisualizer viewer("Patches");
		//viewer.addPointCloud(colored_cloud);
		viewer.addPointCloud(cloud_deb);
		viewer.spin();  
	}
  return clusters;
}


void vision::extract_features_patches(map<int,CloudPtr> segs,map<int,PointCloud<pcl::Normal>::Ptr> seg_normals,map<int,set<int> > segs_nns,map<int,vec> &hist_sis,map<int,vec> &hist_cvs,map<int,vec> &fet_si,map<int,vec> &fet_cv,map<int,vec> &fet_si_max,map<int,vec> &fet_cv_max)
{ 
	hist_normals_.clear();
	hist_cvts_.clear();
	cvts_all_.clear();
	normals_all_.clear();

	search::KdTree<PointType>::Ptr tree (new search::KdTree<PointType>);
  vector<int> k_indices;
  vector<float> k_sqr_distances;
	double radius = 0.01;
	int cids = 0;
	map<int,CloudPtr>::iterator it;
	for(it = segs.begin(); it != segs.end(); it++)
	{
		//cout << (*it).first << " : " << segs.size()  << endl;
		int ref_id = (*it).first;
		set<int> cur_nns = segs_nns[ref_id];
		//cout << "#of nns: " << cur_nns.size() << endl;
		set<int>::iterator its;
		CloudPtr cloud_cur(new Cloud);
		PointCloud<pcl::Normal>::Ptr normals_cur(new PointCloud<pcl::Normal>);
		vector<double> si_s;
		vector<double> curv_s;

    vector<vec> normal_s;
		vector<double> normal_sx;
		vector<double> normal_sy;
		vector<double> normal_sz;
    vector<double> cvt_s;
		for(size_t ip = 0; ip < segs[ref_id]->points.size(); ip++)
			cloud_cur->points.push_back(segs[ref_id]->points[ip]);
		for(size_t ip = 0; ip < seg_normals[ref_id]->points.size(); ip++)
			normals_cur->points.push_back(seg_normals[ref_id]->points[ip]);

		for(its = cur_nns.begin(); its != cur_nns.end(); its++)
		{
			CloudPtr nn_cloud = segs[*its];	
			PointCloud<pcl::Normal>::Ptr nns_normals = seg_normals[*its];
			for(size_t ip = 0; ip < nn_cloud->points.size(); ip++)
				cloud_cur->points.push_back(nn_cloud->points[ip]);
			for(size_t ip = 0; ip < nns_normals->points.size(); ip++)
				normals_cur->points.push_back(nns_normals->points[ip]);
		}
		tree->setInputCloud(cloud_cur);
		for( size_t ip = 0; ip < segs[ref_id]->points.size(); ip++ )
		{
			tree->radiusSearch(*cloud_cur,ip,radius,k_indices,k_sqr_distances);
			bool found = false;
			for( size_t ik = 0;( (ik < k_indices.size()) && (!found) ); ik++ )
			{
				if( k_indices[ik] >= segs[ref_id]->points.size() )				
					found = true;
			}
			pcl::PrincipalCurvaturesEstimation<PointType,pcl::Normal,pcl::PrincipalCurvatures> cvt;	
      float pcx, pcy, pcz, pc1, pc2;
      cvt.computePointPrincipalCurvatures(*normals_cur,ip,k_indices,pcx,pcy,pcz,pc1,pc2);			
			double shape_index = (-2./M_PI) * atan((pc1 + pc2) / (pc1 - pc2) );
			double curv = sqrt( (pow(pc1,2)+pow(pc2,2)) / 2.);
			si_s.push_back(shape_index);
			curv_s.push_back(shape_index);

			vec cur_n(3);
			cur_n(0) = normals_cur->points[ip].normal_x;
			cur_n(1) = normals_cur->points[ip].normal_y;
			cur_n(2) = normals_cur->points[ip].normal_z;
			normal_s.push_back(cur_n);
			normal_sx.push_back(cur_n(0));
			normal_sy.push_back(cur_n(1));
			normal_sz.push_back(cur_n(2));

			NormalEstimation<PointType, pcl::Normal> ne;
			ne.setInputCloud(cloud_cur);
			ne.setSearchMethod(tree);
			ne.setRadiusSearch (0.01);
			float nx,ny,nz,cvt_;
			ne.computePointNormal(*cloud_cur,k_indices,nx,ny,nz,cvt_);
			cvt_s.push_back(cvt_);

			cids++;
		}
		vec cur_fet_si,cur_fet_cv,cur_fet_si_max,cur_fet_cv_max;
		//cout << "to compute single patch feature" << endl;
		fet_si[ref_id] = cur_fet_si;
		fet_cv[ref_id] = cur_fet_cv;
		fet_si_max[ref_id] = cur_fet_si_max;
		fet_cv_max[ref_id] = cur_fet_cv_max;
		//cout << "to make histogram" << endl;
		vec hist_cvt = make_histogram(cvt_s);
		hist_cvts_[ref_id] = hist_cvt;
		
		vec hist_nx = make_histogram(normal_sx);
		vec hist_ny = make_histogram(normal_sy);
		vec hist_nz = make_histogram(normal_sz);
		Mat<double> hist_n(3,hist_nx.n_elem);
		for(int ie = 0; ie < hist_nx.n_elem; ie++)
		{
			hist_n(0,ie) = hist_nx(ie);
			hist_n(1,ie) = hist_ny(ie);
			hist_n(2,ie) = hist_nz(ie);
		}
		hist_normals_[ref_id] = hist_n;
		
		cvts_all_[ref_id] = cvt_s;
		normals_all_[ref_id] = normal_s;

		vec hist_si = make_histogram(si_s);
		vec hist_curv = make_histogram(curv_s);
		hist_sis[ref_id] = hist_si;
		hist_cvs[ref_id] = hist_curv;
		
//		cout << "shape index: " << endl << hist_si << endl;
//		cout << "curvedness: " << endl << hist_curv << endl;
	}
//	cout << "#of survivals: " << cids << endl;

}

vec vision::make_histogram(vector<double> fet,double min_value,double max_value)
{
	double min_val = FLT_MAX;
	double max_val = -FLT_MAX;

	for(size_t i = 0; i < fet.size(); i++)
	{
		if(fet[i] < min_val)
			min_val = fet[i];
		if(fet[i] > max_val)
			max_val =fet[i];
	}

	vec hist(bin_size_);
	hist.fill(0);
	for(size_t i = 0; i < fet.size(); i++)
	{
		double val = fet[i];
		// y = y0 + (y1 -y0) * (x - x0) / (x1 - x0);
		int id = 0. + ((double)bin_size_-1 - 0.) * (val - min_val) / (max_val - min_val);		
		if( (id < 0 ) || (id >= bin_size_) )
			continue;
		hist(id) ++;
	}
	return hist;
}

void vision::aggregate_part_features(map<int,vector<int> > p_ids,vector<int> aff_ids,map<int,vec> hist_sis,map<int,vec> hist_cvs,map<int,vec> fet_si,map<int,vec> fet_cv,map<int,vec> fet_si_max,map<int,vec> fet_cv_max,map<int,CloudPtr> segment_voxels,map<int,set<int> > segs_nns)
{
//	string path_fets_txt;
//	path_fets_txt = default_path_ + "patches-pcl-fg-txt/";
  boost::filesystem::path fet_path_txt(path_fets_txt);
  boost::filesystem::create_directories(fet_path_txt);

//  string parts_path;
//	parts_path = default_path_ + "parts-pcd-fg/";
  boost::filesystem::path parts_path_dir(parts_path);
  boost::filesystem::create_directories(parts_path_dir);

  string obj_path = parts_path + file_name_ + "/";
  boost::filesystem::path obj_path_dir(obj_path);
  boost::filesystem::create_directories(obj_path_dir);
 
  string nn_file_name = obj_path + "segs.txt";
//  cout << "to save in: " << nn_file_name << endl;
  ofstream ofs(nn_file_name.c_str());

	map<int,CloudPtr>::iterator itc;
	for(itc = segment_voxels.begin(); itc != segment_voxels.end(); itc++)
	{
		ofs << (*itc).first;
		if(segs_nns.find((*itc).first) == segs_nns.end()) 
    	ofs << endl;		
		else
		{
    	set<int>::iterator its;
			set<int> nns = segs_nns[(*itc).first];
    	for(its = nns.begin(); its !=nns.end(); its++)
      	ofs << " " << (*its);
    	ofs << endl;			
		}
	}

  map<int,set<int> >::iterator itn;
 /* for(itn = segs_nns.begin(); itn != segs_nns.end(); itn++)
  {
    ofs << (*itn).first;
    set<int> nns = (*itn).second;
    set<int>::iterator its;
    for(its = nns.begin(); its !=nns.end(); its++)
      ofs << " " << (*its);
    ofs << endl;
  }
*/
  ofs.close();
		
	map<int,vector<int> >::iterator it;
	for(it = p_ids.begin(); it != p_ids.end(); it++)
	{
		int p_id = (*it).first;
		vector<int> segs_ids = (*it).second;
		vector<vec> si_s;
		vector<vec> cv_s;

		stringstream ss;
		ss << p_id;
    string p_id_str = ss.str();
	/*	string cur_path = dir_name_ + "part" + ss.str() + "/";
		boost::filesystem::path dir_path(cur_path);
		boost::filesystem::create_directories(dir_path);
*/
		for(size_t i = 0; i < segs_ids.size(); i++)
		{
			int s_id = segs_ids[i];
			si_s.push_back(hist_sis[s_id]);
			cv_s.push_back(hist_cvs[s_id]);	
			
			stringstream ss_s;
			ss_s << i;
			string id_s = ss_s.str();
			stringstream ss_id;
			ss_id << id_s;
      string si_name,cv_name,sim_name,cvm_name,root_name;

			stringstream ss_a;
			ss_a << aff_ids[p_id];
			string sa = ss_a.str();
			stringstream ss_t;
			ss_t << s_id;
			string st = ss_t.str();

      root_name = file_name_ + "_" + sa + "_"  + st + "_sim";
      si_name = path_fets_txt + root_name + ".txt";
      fet_si[s_id].save(si_name,arma_ascii);

      root_name = file_name_ + "_" + sa + "_"  + st + "_cvm";
      cv_name = path_fets_txt + root_name + ".txt"; 
      fet_cv[s_id].save(cv_name,arma_ascii);

      root_name = file_name_ + "_" + sa + "_"  + st + "_simax"; 
      sim_name = path_fets_txt + root_name + ".txt";
      fet_si_max[s_id].save(sim_name,arma_ascii);

      root_name = file_name_ + "_" + sa + "_"  + st + "_cvmax";
      cvm_name = path_fets_txt + root_name + ".txt";
      fet_cv_max[s_id].save(cvm_name,arma_ascii);


			// ** new features ** //
      root_name = file_name_ + "_" + sa + "_"  + st + "_histn";
      si_name = path_fets_txt + root_name + ".txt";
      hist_normals_[s_id].save(si_name,arma_ascii);

      root_name = file_name_ + "_" + sa + "_"  + st + "_histc";
      si_name = path_fets_txt + root_name + ".txt";
      hist_cvts_[s_id].save(si_name,arma_ascii);

			vec cvt_all(cvts_all_[s_id].size());	
			cvt_all.fill(0);
			for(size_t ie = 0; ie < cvts_all_[s_id].size(); ie++)
				cvt_all(ie) = cvts_all_[s_id][ie];
      root_name = file_name_ + "_" + sa + "_"  + st + "_curv";
      si_name = path_fets_txt + root_name + ".txt";
      cvt_all.save(si_name,arma_ascii);

			Mat<double> normal_all(normals_all_[s_id].size(),3);
			normal_all.fill(0);
			for(size_t ie = 0; ie < normals_all_[s_id].size(); ie++)
			{	
				normal_all(ie,0) = normals_all_[s_id][ie](0);
				normal_all(ie,1) = normals_all_[s_id][ie](1);
				normal_all(ie,2) = normals_all_[s_id][ie](2);
			}						
      root_name = file_name_ + "_" + sa + "_"  + st + "_normal";
      si_name = path_fets_txt + root_name + ".txt";
      normal_all.save(si_name,arma_ascii);
			// ** //

//			string cloud_name = cur_path + "cls" + id_s + ".pcd";
      string cloud_name = obj_path + "cls" + "_" + p_id_str + "_" + st + ".pcd";
 //     cout << "to save cloud in: " << cloud_name << endl;
			CloudPtr cur_cloud = segment_voxels[s_id];
			cur_cloud->width = 1;
			cur_cloud->height = cur_cloud->points.size();
			pcl::io::savePCDFileASCII(cloud_name,*cur_cloud);
		}	
	}
	
}
