/*<vision.h: Header files for object parts' feature extraction>
    Copyright (C) <2017>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _VISION_H_
#define _VISION_H_

#define  ARMA_DONT_USE_WRAPPER
#include <fstream>
#include <string>
#include <math.h>
#include <float.h>
#include <armadillo>

#include <pcl/registration/transformation_estimation.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/cloud_viewer.h>
#include <pcl/features/normal_3d.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/registration/ia_ransac.h>
#include <pcl/features/principal_curvatures.h>
#include <pcl/segmentation/region_growing.h>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/filters/passthrough.h>
#include <pcl/filters/voxel_grid.h>

typedef pcl::PointXYZ PointType;
typedef pcl::PointCloud<PointType> Cloud;
typedef typename Cloud::Ptr CloudPtr;
typedef typename Cloud::ConstPtr CloudConstPtr;

typedef pcl::PointXYZRGB ColorPointType;
typedef pcl::PointCloud<ColorPointType> ColorCloud;
typedef typename ColorCloud::Ptr ColorCloudPtr;
typedef typename ColorCloud::ConstPtr ColorCloudConstPtr;

typedef pcl::PointXYZL LPointType;
typedef pcl::PointCloud<LPointType> LCloud;
typedef typename LCloud::Ptr LCloudPtr;

using namespace pcl;
using namespace std;
using namespace arma;

const int bin_size_ = 10;
const double smooth_thresh_ = 3.0;
// Enable for visualization if true
const bool debug_ = false;

class vision
{
  public:
		string dir_name_;
		string file_name_;
		string path_fets_txt,parts_path,files_path_;

		map<int,Mat<double> > hist_normals_;
		map<int,vec> hist_cvts_;
		map<int,vector<double> > cvts_all_;
		map<int,vector<vec> > normals_all_;

    vision();
    ~vision();	
    CloudPtr load_cloud(string file_name);
		vector<string> get_all_files(string dir_name,string pattern=".pcd");
		CloudPtr segment_pcd_euc(CloudPtr cloud);
		vector<CloudPtr> read_pcd_parts(string dir_name,CloudPtr &cloud_orig,vector<int> &aff_ids);
    void learn_parts_features_pcl(string d_path = "");
		void learn_parts_features_pcl_test_planar_seg();
    void extract_between_patch_features();
    double compute_curvature_between_patches(CloudPtr cloud1,CloudPtr cloud2);

    PointCloud<pcl::Normal>::Ptr estimate_normal(CloudPtr cloud);
    vector<pcl::PointIndices> get_regions_from_pcd(CloudPtr cloud,PointCloud<pcl::Normal>::Ptr normals,map<int,CloudPtr> &segment_voxels,map<int,PointCloud<pcl::Normal>::Ptr> &segment_normals,map<int,set<int> > &segs_nns,double seg_thresh=-1);
		CloudPtr segment_pcd_euc_iter(CloudPtr cloud);

		void extract_features_patches(map<int,CloudPtr> segs,map<int,PointCloud<pcl::Normal>::Ptr> seg_normals,map<int,set<int> > segs_nns,map<int,vec> &hist_sis,map<int,vec> &hist_cvs,map<int,vec> &fet_si,map<int,vec> &fet_cv,map<int,vec> &fet_si_max,map<int,vec> &fet_cv_max);
		vec make_histogram(vector<double> fet_ref,double min_val = -1,double max_val = 1);
		map<int,vector<int> > get_segments_parts_ids(map<int,CloudPtr> segment_voxels,vector<CloudPtr> parts,CloudPtr cloud);
		void aggregate_part_features(map<int,vector<int> > p_ids,vector<int> aff_ids,map<int,vec> hist_sis,map<int,vec> hist_cvs,map<int,vec> fet_si,map<int,vec> fet_cv,map<int,vec> fet_si_max,map<int,vec> fet_cv_max,map<int,CloudPtr> segment_voxels,map<int,set<int> > segs_nns);

		void learn_parts_features_pcl_test();
		void save_patch_test_features(map<int,vec> hist_sis,map<int,vec> hist_cvs,map<int,vec> fet_si,map<int,vec> fet_cv,map<int,vec> fet_si_max,map<int,vec> fet_cv_max,map<int,CloudPtr> segment_voxels,map<int,set<int> > segs_nns);
};

#endif
