/*<vision.cpp: Functions for object parts visualization after segmentation>
    Copyright (C) <2017>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "vision.h"

vision::vision()
{
}

vision::~vision()
{
}

CloudPtr vision::load_cloud(string file_name)
{
  pcl::PCDReader reader;
  CloudPtr cloud (new Cloud);
  reader.read (file_name, *cloud);
  return cloud;
}

PointCloud<pcl::Normal>::Ptr vision::estimate_normal(CloudPtr cloud)
{
  NormalEstimation<PointType, pcl::Normal> ne;
  ne.setInputCloud(cloud);
  pcl::search::KdTree<PointType>::Ptr tree (new pcl::search::KdTree<PointType> ());
  ne.setSearchMethod(tree);
  PointCloud<pcl::Normal>::Ptr cloud_normals (new PointCloud<pcl::Normal>);
//  ne.setKSearch(30);
  ne.setRadiusSearch(normal_rad_);
  ne.setSearchSurface(cloud);
  ne.compute (*cloud_normals);
  return cloud_normals;
}

vector<string> vision::get_all_files(string dir_name,string pattern)
{
	int min_length = 1;
  vector<string> files;
  DIR *dir;
  struct dirent *ent;
  if ((dir = opendir (dir_name.c_str())) != NULL) 
  {
    while ((ent = readdir (dir)) != NULL) {
      string name = ent->d_name;
			if( name.length() < min_length )
				continue;
      if( name.find(pattern) == string::npos )
        continue;
      string file_name = dir_name + name;
      files.push_back(file_name);
    }
    closedir (dir);
  } 
 // else
   //  cout << "can't open dir: " << dir_name <<endl;

  return files;
}

void vision::store_parts_normals_evals()
{
  boost::filesystem::path eval_path(store_path);
  boost::filesystem::create_directories(store_path);
	
	vector<string> dir_all = get_all_files(eval_dir,"");			
  for(size_t i = 0; i < dir_all.size(); i++)
//	for(size_t i = 0; i < 10; i++)
  {
    if(dir_all[i].find(".") != string::npos)
      continue;
		ColorCloudPtr cloud_deb(new ColorCloud);
    cout << dir_all[i] << endl;
    string root_name = dir_all[i].substr(dir_all[i].find_last_of("/")+1);
    root_name = root_name.substr(0,root_name.find_last_of("_"));

    string f_name = dir_all[i].substr(dir_all[i].find_last_of("/")+1);
		string cur_dir_path = store_path + f_name + "/";

    boost::filesystem::path cur_path(cur_dir_path);
    boost::filesystem::create_directories(cur_path);

    string part_txt_file = dir_all[i] + "/" + "parts.txt";
//		cout << part_txt_file << endl;
    ifstream ifs(part_txt_file.c_str());
    int max_parts = 0;
    vector<int> patch_ids;
		vector<int> part_ids;
    while(ifs.good())
    {
      string line;
      getline(ifs,line);
			//cout << line << endl;
      if(line.length() == 0)
        continue;
			int part_id = max_parts;
			max_parts++;
			while(line.find(" ") != string::npos)
			{
				string patch_str = line.substr(0,line.find_first_of(" "));
				line = line.substr(line.find_first_of(" ")+1);
				int patch_id;
				stringstream ss_patch(patch_str);
				ss_patch >> patch_id;
				patch_ids.push_back(patch_id-1);
				part_ids.push_back(part_id);
			}
			int patch_id;
			stringstream ss_patch(line);			
			ss_patch >> patch_id;
			patch_ids.push_back(patch_id-1);
			part_ids.push_back(part_id);
    }
	//	cout << endl << endl;
    vector<CloudPtr> clouds(max_parts);
//		cout << "max parts: " << max_parts << endl;
    for(size_t j = 0; j < clouds.size(); j++)
    {
      CloudPtr cloud(new Cloud);
      clouds[j] = cloud;
    }				
		
		string patch_orig_pcds_dir = base_dir + f_name + "/";
		vector<string> patch_all = get_all_files(patch_orig_pcds_dir,".pcd");
	//	cout << "#of patches " << patch_all.size() << endl;
		int num_rej = 0;
		vector<CloudPtr> patch_orig_clouds(patch_all.size());
		for(size_t j = 0; j < patch_all.size(); j++)
		{
			CloudPtr cur_cloud(new Cloud);
			patch_orig_clouds[j] = cur_cloud;
		}
		bool valid = true;
		for(size_t j = 0; j < patch_all.size(); j++)
		{
			string cur_p_name = patch_all[j].substr(patch_all[j].find_last_of("/")+1);		
			cur_p_name = cur_p_name.substr(0,cur_p_name.find("."));
			cur_p_name = cur_p_name.substr(cur_p_name.find_last_of("_")+1);
			stringstream ss(cur_p_name);
			int patch_id;
			ss >> patch_id;
			
			if(patch_id >= patch_orig_clouds.size())
			{
				valid = false;
				continue;
			}
//			cout << patch_all[j] << " : " << patch_id << endl;
			CloudPtr cur_cloud = load_cloud(patch_all[j]);
//			cout << cur_cloud->points.size() << endl;		
			patch_orig_clouds[patch_id] = cur_cloud;
		}
		if(!valid)
			continue;
    for(size_t j = 0; j < patch_ids.size(); j++)
    {
      int patch_id = patch_ids[j];
      int part_id = part_ids[j];

			if(patch_id >= patch_orig_clouds.size() ) 
			{
				valid = false;
				continue;
			}
			//cout << patch_id << " : " << part_id << endl;

//			cout << patch_id << " : " << patch_orig_clouds.size() << endl;
      CloudPtr cur_cloud = patch_orig_clouds[patch_id];
		//	cout << cur_cloud->points.size() << endl;
      if(cur_cloud->points.size() == 0)
			{
				num_rej ++;
        continue;
			}
      CloudPtr cloud_orig(new Cloud);
      *cloud_orig = *clouds[part_id];
      for(size_t ip = 0; ip < cur_cloud->points.size(); ip++)
        cloud_orig->points.push_back(cur_cloud->points[ip]);
      *clouds[part_id] = *cloud_orig;
    }
		if(!valid)
			continue;
//		cout << "#of rejected: " << num_rej << endl;
//		cout << "to generate parts" << endl;
		num_rej = 0;
	//	ColorCloudPtr cloud_deb(new ColorCloud);
    for(size_t j = 0; j < clouds.size(); j++)
    {
      if(clouds[j]->points.size() == 0)
			{
				num_rej ++;
        continue;
			}
			double r = rand() % 255;
			double g = rand() % 255;
			double b = rand() % 255;
//			cout << "#of rejected: " << num_rej << endl;
      PointCloud<pcl::Normal>::Ptr cur_normal = estimate_normal(clouds[j]);
      pcl::PointCloud<PointNormal>::Ptr cloud_n_xyz(new pcl::PointCloud<PointNormal>);
      for(size_t ip = 0; ip < clouds[j]->points.size(); ip++)
      {
        PointNormal pn;
        pn.x = clouds[j]->points[ip].x; pn.y = clouds[j]->points[ip].y; pn.z = clouds[j]->points[ip].z;
        pn.normal_x = cur_normal->points[ip].normal_x;
        pn.normal_y = cur_normal->points[ip].normal_y;
        pn.normal_z = cur_normal->points[ip].normal_z;
        pn.curvature = cur_normal->points[ip].curvature;
        cloud_n_xyz->points.push_back(pn);
				
				ColorPointType pc;
				pc.x = pn.x; pc.y = pn.y; pc.z = pn.z;
				pc.r = r; pc.g = g; pc.b = b;
				cloud_deb->points.push_back(pc);
      }
      stringstream ss_p;
      ss_p << cur_dir_path << "part" << j << ".pcd";
      string cur_pcd_name = ss_p.str();
//			cout << "to save in: " << cur_pcd_name << endl;
      cloud_n_xyz->width = 1;
      cloud_n_xyz->height = cloud_n_xyz->points.size();
      pcl::io::savePCDFileASCII(cur_pcd_name,*cloud_n_xyz);
    }
		if(debug_)
		{
			pcl::visualization::PCLVisualizer viewer("Segmented Parts");
			viewer.addPointCloud(cloud_deb);
			viewer.spin();  
		}				
	}	

}
