/*<vision.h: Header files for object parts visualization>
    Copyright (C) <2017>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _VISION_H_
#define _VISION_H_
#include <fstream>
#include <string>
#include <math.h>
#include <float.h>

#include <pcl/registration/transformation_estimation.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/octree/octree.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/cloud_viewer.h>
#include <pcl/features/normal_3d.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/registration/ia_ransac.h>
#include <pcl/segmentation/supervoxel_clustering.h>
#include <pcl/segmentation/region_growing.h>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/registration/icp.h>

typedef pcl::PointXYZ PointType;
typedef pcl::PointCloud<PointType> Cloud;
typedef typename Cloud::Ptr CloudPtr;
typedef typename Cloud::ConstPtr CloudConstPtr;

typedef pcl::PointXYZRGB ColorPointType;
typedef pcl::PointCloud<ColorPointType> ColorCloud;
typedef typename ColorCloud::Ptr ColorCloudPtr;
typedef typename ColorCloud::ConstPtr ColorCloudConstPtr;

using namespace pcl;
using namespace std;

// Enable visualization if true
const bool debug_ = false;
const double normal_rad_ = 0.01; // 0.03

class vision
{
	public:
		string base_dir,store_path,eval_dir;
		
		vision();
    ~vision();

		vector<string> get_all_files(string dir_name,string pattern=".pcd");
		CloudPtr load_cloud(string file_name);
		PointCloud<pcl::Normal>::Ptr estimate_normal(CloudPtr cloud);
		void store_parts_normals_evals();
};
#endif
