% <learn_part_features.m: Learning features of the parts based on patches>
%     Copyright (C) <2017>  <Safoura Rezapour Lakani>
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%
function hists = learn_part_features(train_parts,patch_names,patch_fets,patch_cls,id_t,config)

hists = get_parts_histogram(train_parts,patch_names,patch_fets,patch_cls,id_t,config);


%%
function hists = get_parts_histogram(train_parts,patch_names,patch_fets,patch_cls,id_t,config)

%path = '/home/c7031016/Desktop/safoura-stuff/3dv-codes/data/';
path = config.data_path;

cid = 1;
for i = 1:numel(train_parts.file_names)
    f_name = train_parts.file_names{i}
    pts = give_patches(f_name,patch_names,patch_fets);
    part_ids = get_patch_part_ids(train_parts.parts{i});
    cls_ids = get_patch_cls_id(patch_cls,pts.fets');
    num_patches = size(patch_cls,1);
    for j = 1:numel(part_ids)
        hist = get_patch_histogram(part_ids{j},cls_ids,pts.ids,num_patches);
        hists(cid,:) = hist;
        cid = cid+1;
    end
%     if(i == 3)
%         break;
%     end
end

if(numel(id_t) > 0)
    
    if(id_t == 1)
        %hist_name = strcat(path,'parts_hists_pcd_novel.mat');
        hist_name = strcat(path,config.train_part_hist);
    else
        %  hist_name = strcat(path,'parts_hists_test_pcd_novel.mat');
        hist_name = strcat(path,config.test_part_hist);
    end
    save(hist_name,'hists');
end
%%
function pts = give_patches(f_name,patch_names,patch_fets)

cid = 1;
for i = 1:numel(patch_names)
    cur_id_str = patch_names(i).name;
    t = findstr(cur_id_str,f_name);
    if(numel(t) == 0)
        continue;
    end
    t = findstr(patch_names(i).name,'_');
    cur_id_str = cur_id_str(t(numel(t)-1)+1:t(numel(t))-1);
    cur_id = str2num(cur_id_str);
    pts.ids(cid) = cur_id;
    pts.fets(:,cid) = patch_fets(:,i);
    cid = cid +1;
end
%%
function hist = get_patch_histogram(part,cls_ids,patch_ids,num_patches)

hist = zeros(1,num_patches);
for i = 1:numel(part)
    cur_id = part(i);
    patch_id = find(patch_ids == cur_id);
    cl_id = cls_ids(patch_id);
    hist(cl_id) = hist(cl_id) + 1;
end
%%
function p_ids = get_patch_part_ids(part)

parts_ref = part;
max_parts = max(parts_ref(:,1))+1;
segs_ids = ones(max_parts);
p_ids = [];
for j = 1:size(parts_ref,1)
    id = parts_ref(j,1);
    p_ids{id+1}(segs_ids(id+1)) = parts_ref(j,2);
    segs_ids(id+1) = segs_ids(id+1) + 1;
end
%%
function cls_ids = get_patch_cls_id(patch_cls,patch_fets)

%[D,I] = pdist2(means,v1','euclidean','Smallest',1);

%[dists,cls_ids] = pdist2(patch_cls,patch_fets','euclidean','Smallest',1);
[dists,cls_ids] = pdist2(patch_cls,patch_fets,'euclidean','Smallest',1);
