% <learn_script.m: Script for learning part segmentation from ground-truth data>
%     Copyright (C) <2017>  <Safoura Rezapour Lakani>
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%
path = config.data_path;


%%
% call c++ feature extraction
ref_path = strcat(pwd,'/code-extract-features/build/');
fet_ext = strcat(ref_path,'vis'); 
data_path = strcat(config.orig_path,config.train_files_path);
data_patch_path = strcat(config.orig_path,config.train_patch_path);
data_parts_path = strcat(config.orig_path,config.train_part_path);
fet_ext_command = strcat(fet_ext,{' '},'train',{' '},data_path,{' '},data_patch_path,{' '},data_parts_path);
fet_ext_command_str = fet_ext_command{1}
[status,cmdout] = system(fet_ext_command_str,'-echo');

%%

fprintf('to get train patches\n');
trains = get_patches(config,1);

fprintf('to get train parts\n');
trains_parts = get_parts(config,1);

train_parts_str = strcat(path,config.train_parts_mat);
tests_parts = importdata(train_parts_str);

fprintf('to learn patch clusters\n');
patch_cls = patch_clustering(trains,[],config);

fprintf('to learn training part features\n');
parts_hists = learn_part_features(trains_parts,trains.files_names,trains.histn,patch_cls,1,config);

fprintf('to learn part clusters\n');
[id_cls,part_cls] = part_clustering(parts_hists,config);

fprintf('to collect part statistics\n');
collect_statistics_parts(trains_parts,trains.files_names,trains.histn,parts_hists,part_cls,patch_cls,config);

fprintf('to learn curvature training data\n');
[p_trs_train,l_trs_train] = train_curvature_classifier(trains_parts);

fprintf('to learn curvature test data\n');
[p_trs_test,l_trs_test] = train_curvature_classifier(tests_parts);

fprintf('to learn curvature classifier\n');
script_classify(p_trs_train,l_trs_train,p_trs_test,l_trs_test,config);

Mdl_str = strcat(path,'rbf-',config.svm_cvt_dir,'mdl.mat');
Mdl = importdata(Mdl_str);

% if the same its 1 otherwise its 2
t = strfind(config.mdl_cvt,'/');
p_svms_path = config.mdl_cvt(1:t(1));
p_svms_path = strcat(path,p_svms_path,'prob_svms');
prob_svms = fitPosterior(Mdl.BinaryLearners{1},p_trs_train',l_trs_train');
save(p_svms_path,'prob_svms');

p_occ_prob_str = strcat(path,config.occ_prob_mat);
p_occ_prob = importdata(p_occ_prob_str);

w = mrf_train_demo(trains_parts,trains.files_names,trains.histn,patch_cls,part_cls,p_occ_prob,prob_svms,config);
