% <mrf_test_all.m: Provides nodes and adjacency information for MRF based segmentation and stores the predicted labels.>
%     Copyright (C) <2017>  <Safoura Rezapour Lakani>
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%
function [parts_all] = mrf_test_all(test_parts,test_patch_names,test_patch_fets,patch_cls,part_cls,p_occ_prob,w,Mdl,config)

parts_all = [];
overlap_all = 0;
nStates = size(part_cls,1);

path = config.orig_path;
dir_name = strcat(path,config.parts_save_path);

mkdir(dir_name);

[w_unary,w_pair] = make_weights_matrices(w,nStates);

for i = 1:numel(test_parts.segs)
%for i = 2000:2005
    fprintf('to compute parts for %s \n',test_parts.file_names{i});
    
    patches_adj = test_parts.segs{i};
    if(numel(patches_adj) == 0)
        continue;
    end
    
    %disp(patches_adj)
    f_name = test_parts.file_names{i};
    pts = give_patches(f_name,test_patch_names,test_patch_fets);
    %fprintf('%d\n',numel(pts));
    if(numel(pts) == 0)
        continue;
    end

    cur_dir_name = strcat(dir_name,test_parts.file_names{i},'/');
    mkdir(cur_dir_name);
    fprintf('to save parts %s\n',cur_dir_name);
    
    cls_ids = get_patch_cls_id(patch_cls,pts.fets');
    
    [optimalDecoding] = mrf_test_cvt(patches_adj,pts,test_parts.cvts{i},cls_ids,p_occ_prob,nStates,w_unary,w_pair,Mdl);
    parts_all.parts{i} = optimalDecoding;
    parts_all.file_names{i} = test_parts.file_names{i};
    
    %parts_ref = test_parts.parts{i};
    %max_parts = max(parts_ref(:,1))+1;
    %segs_ids = ones(max_parts);
    %p_ids = [];
    %for j = 1:size(parts_ref,1)
     %   id = parts_ref(j,1);
      %  p_ids{id+1}(segs_ids(id+1)) = parts_ref(j,2)+1;
       % segs_ids(id+1) = segs_ids(id+1) + 1;
    %end
    %   unique(optimalDecoding);
    
    labels = get_unique_labels(patches_adj,optimalDecoding);
    
    
    %max_parts_t = max(optimalDecoding);
    max_parts_t = max(labels);
    segs_ids_t = ones(max_parts_t);
    p_ids_t = [];
    %for j = 1:numel(optimalDecoding)
    for j = 1:numel(labels)    
       % id = optimalDecoding(j);
        id = labels(j);
        p_ids_t{id}(segs_ids_t(id)) = j;
        segs_ids_t(id) = segs_ids_t(id) + 1;
    end
    
    part_name = strcat(cur_dir_name,'parts.mat');
    save(part_name,'p_ids_t');
    part_name_txt = strcat(cur_dir_name,'parts.txt');
    ftemp = fopen(part_name_txt,'w' );
    for j = 1:numel(p_ids_t)
        for k = 1:numel(p_ids_t{j})
            fprintf(ftemp, '%d',p_ids_t{j}(k));
            if(k == numel(p_ids_t{j}))
                fprintf(ftemp, '\n');
            else
                fprintf(ftemp, ' ');
            end
        end
    end
    fclose(ftemp);
end
% parts_all_name = strcat(path,config.segmented_parts_mat);
% save(parts_all_name,'parts_all');
%%
function labels = get_unique_labels(patches_adj,optimalDecoding)
labels = [];
nNodes = numel(patches_adj);
adj = zeros(nNodes,nNodes);
for k =1:numel(patches_adj)
    a = patches_adj{k};
    if(numel(a) < 2)
        continue;
    end
    for j =2:numel(a)
        adj(a(1)+1,a(j)+1) = 1;
    end
end

l_max = max(optimalDecoding);
un_labels = unique(optimalDecoding);

part_ids = [];
for i = 1:numel(un_labels)
    cur_l = un_labels(i);
    cur_ids = find(optimalDecoding == cur_l);
    part_ids{i} = cur_ids;
end

labels = zeros(1,numel(optimalDecoding));

l_id = 1;

for i = 1:numel(part_ids)
    a = part_ids{i};
    cur_l = zeros(1,numel(a));
    for id = 1:numel(a)
        if(cur_l(id) == 0)
            cur_l(id) = id;
        end
        for j = id+1:numel(a)
            if(adj(a(id),a(j)) == 1)  
                cur_l(j) = cur_l(id); 
            end
        end       
    end
    
    num_regs = unique(cur_l);
    for j = 1:numel(num_regs)
        cur_ids = find(cur_l == num_regs(j));
        for k = 1:numel(cur_ids)
            labels(a(cur_ids(k))) = l_id;
        end        
        l_id = l_id+1;
    end
    
end

%optimalDecoding

%labels

%%
function [w_unary,w_pair] = make_weights_matrices(w,nStates)

for i = 1:nStates
    w_unary(i) = w(i);
end

w_rest = w(nStates+1:end);
w_pair = zeros(nStates,nStates);
cid = 1;
for i = 1:nStates
    for j = i+1:nStates
        w_pair(i,j) = w_rest(cid);
        w_pair(j,i) = w_rest(cid);
        cid = cid+1;
    end
end
%%
function cls_ids = get_patch_cls_id(patch_cls,patch_fets)

%[D,I] = pdist2(means,v1','euclidean','Smallest',1);

%[dists,cls_ids] = pdist2(patch_cls,patch_fets','euclidean','Smallest',1);
[dists,cls_ids] = pdist2(patch_cls,patch_fets,'euclidean','Smallest',1);
%%
function pts = give_patches(f_name,patch_names,patch_fets)

cid = 1;
pts = [];
for i = 1:numel(patch_names)
  %  f_name
    cur_id_str = patch_names(i).name;
    t = findstr(cur_id_str,f_name);
    if(numel(t) == 0)
        continue;
    end
    %patch_names(i)
    t = findstr(patch_names(i).name,'_');
    cur_id_str = cur_id_str(t(numel(t)-1)+1:t(numel(t))-1);
    cur_id = str2num(cur_id_str);
    pts.ids(cid) = cur_id;
    pts.fets(:,cid) = patch_fets(:,i);
    cid = cid +1;
end
